/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SWO_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void SurfaceWrapOp::initialize(void)
{
	//* NOTE inherits attributes from SurfaceBindOp

	catalogRemove("Bind Center");

	catalog<String>("BindPrefix","hint")=
			"Prefix of attribute names on input surface describing"
			" where each point on the input surface is attached to"
			" the associated driver";

	catalog<Real>("Distance Bias")=0.0;
	catalog<Real>("Distance Bias","min")=0.0;
	catalog<Real>("Distance Bias","max")=10.0;
	catalog<String>("Distance Bias","hint")=
			"Significance of more distant faces towards weighting";

	catalog<bool>("Debug Bind")=false;
	catalog<bool>("Debug Bind","joined")=true;
	catalog<String>("Debug Bind","hint")=
			"Add lines to the output showing the surface refinement."
			"  These lines are put in a 'debug' group.";

	catalog<I32>("Debug Index")=-1;
	catalog<I32>("Debug Index","min")=-1;
	catalog<I32>("Debug Index","high")=1000;
	catalog<I32>("Debug Index","max")=1000000;
	catalog<String>("Debug Index","enabler")="Debug Bind";
	catalog<String>("Debug Index","hint")=
			"Index of element to examine (-1 to see them all).";

	catalog< sp<Component> >("Deformed Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;

	m_spDebugGroup=new DrawMode();
	m_spDebugGroup->setGroup("debug");
}

void SurfaceWrapOp::handle(Record& a_rSignal)
{
#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle\n");
#endif

	//* access output curves (copy of input surface)
	m_spOutputAccessible=surfaceOutput(a_rSignal);
	if(m_spOutputAccessible.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" output surface not SurfaceAccessibleI\n");
		catalog<String>("error")="can't access output data";
		return;
	}

	if(catalog<BWORD>("Debug Bind"))
	{
		m_spDrawI=drawI(a_rSignal);
		if(!m_spDrawI.isValid())
		{
			feLog("DrawTestOp::handle DrawI invalid\n");
			catalog<String>("error")="invalid drawing interface";
			return;
		}
	}
	else
	{
		m_spDrawI=NULL;
	}

	//* access reference surface
	m_spInputAccessible=catalog< sp<Component> >("Input Surface");
	if(m_spInputAccessible.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" reference surface not SurfaceAccessibleI\n");
		catalog<String>("error")="can't access reference surface";
		return;
	}

	//* access original driver
	m_spDriverAccessible=catalog< sp<Component> >("Driver Surface");
	if(m_spDriverAccessible.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" driver surface not SurfaceAccessibleI\n");
		catalog<String>("error")="can't access driver surface";
		return;
	}

	//* access deformed driver
	m_spDeformedAccessible=catalog< sp<Component> >("Deformed Surface");
	if(m_spDeformedAccessible.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" deformed surface not SurfaceAccessibleI\n");
		catalog<String>("error")="can't access deformed curves";
		return;
	}

	m_spDriver=m_spDriverAccessible->surface();
	if(m_spDriver.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" driver surface failed to provide SurfaceI\n");
		catalog<String>("error")="invalid driver surface";
		return;
	}
	m_spDriver->setRefinement(0);
	m_spDriver->center();

	m_spDeformed=m_spDeformedAccessible->surface();
	if(m_spDeformed.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" deformed surface failed to provide SurfaceI\n");
		catalog<String>("error")="invalid deformed surface";
		return;
	}
	m_spDeformed->setRefinement(0);
	m_spDeformed->center();

	const bool bindPrimitives=catalog<bool>("Bind Primitives");
	m_element=bindPrimitives?
			SurfaceAccessibleI::e_primitive:
			SurfaceAccessibleI::e_point;
	m_attribute=bindPrimitives?
			SurfaceAccessibleI::e_vertices:
			SurfaceAccessibleI::e_position;

	sp<SurfaceAccessorI> spOutputElements=
			m_spOutputAccessible->accessor(m_element,m_attribute);
	FEASSERT(spOutputElements.isValid());

	const U32 elementCount=spOutputElements->count();

	if(m_spDrawI.isValid())
	{
		m_spDrawI->pushDrawMode(m_spDebugGroup);
	}

	U32 runCount=elementCount;
	m_fragmented=catalog<bool>("Bind Fragments");
	m_spFragmentAccessor=NULL;
	if(m_fragmented)
	{
		m_spFragmentAccessor=
			m_spOutputAccessible->accessor(m_element,m_attribute);

		if(m_spFragmentAccessor.isValid())
		{
			const String fragmentAttribute=
					catalog<String>("FragmentAttr");
			m_spFragmentAccessor->fragmentWith(m_element,fragmentAttribute);
			runCount=m_spFragmentAccessor->fragmentCount();
			if(runCount<2)
			{
				catalog<String>("warning")=
						"fragmention failed to divide surface -> reverting";
				m_fragmented=FALSE;
				runCount=elementCount;
			}
		}
		else
		{
			catalog<String>("warning")="fragmention access failed -> ignoring";
			m_fragmented=FALSE;
			runCount=elementCount;
		}
	}

	setFullRange(0,runCount-1);
	OperatorThreaded::handle(a_rSignal);

	if(m_spDrawI.isValid())
	{
		m_spDrawI->popDrawMode();
	}


#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle done\n");
#endif
}

void SurfaceWrapOp::run(I32 a_id,Range a_range)
{
	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;

	const Real distanceBias=catalog<Real>("Distance Bias");
	const I32 debugIndex=catalog<I32>("Debug Index");

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run running thread %d [%d,%d]\n",a_id,start,end);
#endif

	sp<SurfaceAccessorI> spInputElements=
			m_spInputAccessible->accessor(m_element,m_attribute);
	FEASSERT(spInputElements.isValid());

	sp<SurfaceAccessorI> spOutputElements=
			m_spOutputAccessible->accessor(m_element,m_attribute);
	FEASSERT(spOutputElements.isValid());

	std::vector< sp<SurfaceAccessorI> > outputFaces;
	std::vector< sp<SurfaceAccessorI> > outputBarys;
	const I32 bindings=accessBindings(m_spOutputAccessible,
			outputFaces,outputBarys,FALSE);
	if(bindings<1)
	{
		return;
	}

	const bool spiderWrap=catalog<bool>("Spider");
	const BWORD singular=
			(m_fragmented && m_element==SurfaceAccessibleI::e_point);

	sp<SurfaceAccessorI::FilterI> spFilter;

	for(U32 index=start;index<=end;index++)
	{
		I32 filterCount=1;
		U32 pointCount=0;
		if(m_fragmented)
		{
			const String fragment=m_spFragmentAccessor->fragment(index);
			if(!m_spFragmentAccessor->filterWith(fragment,spFilter) ||
					spFilter.isNull())
			{
				continue;
			}
			filterCount=spFilter->filterCount();
			if(filterCount<1)
			{
				continue;
			}
			if(singular)
			{
				pointCount+=filterCount;
			}
			else
			{
				for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
				{
					const U32 filtered=spFilter->filter(filterIndex);
					pointCount+=spInputElements->subCount(filtered);
				}
			}
		}
		else
		{
			pointCount=spInputElements->subCount(index);
		}

		if(pointCount<1)
		{
			continue;
		}

		SpatialVector point[pointCount];
		SpatialVector result[pointCount];
		Real totalWeight[pointCount];

		U32 pointIndex=0;
		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const U32 filtered=
					m_fragmented? spFilter->filter(filterIndex): index;
			const U32 subCount=singular? 1: spInputElements->subCount(filtered);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				point[pointIndex]=
						spInputElements->spatialVector(filtered,subIndex);
				set(result[pointIndex]);
				totalWeight[pointIndex]=0.0;
				pointIndex++;
			}
		}

		const U32 keyIndex=m_fragmented? spFilter->filter(0): index;

		for(U32 binding=0;binding<bindings;binding++)
		{
			const I32 face=outputFaces[binding]->integer(keyIndex);
			if(face<0)
			{
				continue;
			}

			SpatialBary barycenter(0.33,0.33);
			if(!binding && !spiderWrap)
			{
				barycenter=outputBarys[binding]->spatialVector(keyIndex);
			}
			const SpatialTransform xformDriver=
					m_spDriver->sample(face,barycenter);
			const SpatialTransform xformDeformed=
					m_spDeformed->sample(face,barycenter);
			SpatialTransform invDriver;
			invert(invDriver,xformDriver);
			const SpatialTransform conversion=invDriver*xformDeformed;

			const Real distance=
					magnitude(xformDriver.translation()-point[0]);
			const Real weight=1.0/(distance+distanceBias+1e-9);

			for(pointIndex=0;pointIndex<pointCount;pointIndex++)
			{
				SpatialVector transformed;
				transformVector(conversion,point[pointIndex],transformed);
				result[pointIndex]+=weight*transformed;
				totalWeight[pointIndex]+=weight;
			}

			if(m_spDrawI.isValid() && (debugIndex<0 || debugIndex==index))
			{
				const Color yellow(1.0,1.0,0.0,1.0);
				SpatialVector line[2];

				line[0]=point[0];
				line[1]=xformDriver.translation();

				SAFEGUARD;
				m_spDrawI->drawLines(line,NULL,2,DrawI::e_strip,
						false,&yellow);
			}
		}

		pointIndex=0;
		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const U32 filtered=
					m_fragmented? spFilter->filter(filterIndex): index;
			const U32 subCount=singular? 1: spInputElements->subCount(filtered);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				if(totalWeight[pointIndex]>0.0)
				{
					spOutputElements->set(filtered,subIndex,
							result[pointIndex]/totalWeight[pointIndex]);
				}

				pointIndex++;
			}
		}
	}

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run done running thread %d\n",a_id);
#endif
}
